""" Sampler classes for generating in random samples and PDF evaluations. """
from typing import List, Tuple, Union, Callable, Optional
from abc import ABC, abstractmethod, abstractproperty
import warnings
import numpy as np
import scipy.stats
from scipy.integrate import quad
from scipy.special import gamma
import pythia as pt


class Sampler(ABC):
    """Base class for all continuous samplers. """

    # set abstract attributes
    domain: np.ndarray

    @abstractproperty
    def dimension(self) -> int:
        """Dimension of the ambient space."""
        raise NotImplementedError

    @abstractproperty
    def mass(self):
        """Mass of the sampler distribution.

        The integral of the sampler distribution over the domain of
        definition. If the density is normalised this value should be one.
        """
        raise NotImplementedError

    @abstractproperty
    def maximum(self) -> float:
        """Maximum of the pdf."""
        raise NotImplementedError

    @abstractproperty
    def mean(self) -> Union[float, np.ndarray]:
        """Mean value of the pdf."""
        raise NotImplementedError

    @abstractproperty
    def cov(self) -> Union[float, np.ndarray]:
        """(Co)Variance of the pdf."""
        raise NotImplementedError

    @abstractmethod
    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Density of the samplers distribution.

        Computes the density of the samplers underlying distribution at the
        given points `x`.

        Parameters
        ----------
        x : array_like of shape (..., D)
            List of points or single point. `D` is the objects dimension.

        Returns
        -------
        :
            Density values at the points.
        """
        raise NotImplementedError

    @abstractmethod
    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Log-density of the samplers distribution.

        Computes the log-density of the samplers underlying distribution at the
        given points `x`.

        Parameters
        ----------
        x : array_like of shape (..., D)
            List of points or single point. `D` is the objects dimension.

        Returns
        -------
        :
            Log-density values at the points.
        """
        raise NotImplementedError

    @abstractmethod
    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Gradient of log-density of the samplers distribution.

        Computes the gradient of the log-density of the samplers underlying
        distribution at the given points `x`.

        Parameters
        ----------
        x : array_like of shape (..., D)
            List of points or single point. `D` is the objects dimension.

        Returns
        -------
        :
            Gradient values of the log-density at the points with shape
            (..., D).
        """
        raise NotImplementedError

    @abstractmethod
    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Hessian of log-density of the samplers distribution.

        Computes the Hessian of the log-density of the samplers underlying
        distribution at the given points `x`.

        Parameters
        ----------
        x : array_like of shape (..., D)
            List of points or single point. `D` is the objects dimension.

        Returns
        -------
        :
            Hessian values of the log-density at the points with shape
            (..., D, D).
        """
        raise NotImplementedError

    @abstractmethod
    def sample(self, shape: Union[List, Tuple, np.ndarray]) -> np.ndarray:
        """Random values in a given shape.

        Create an array of the given shape and populate it with random samples
        from the samplers distribution.

        Parameters
        ----------
        shape : array_like, optional
            The dimensions of the returned array, should all be positive.
            If no argument is given a single Python float is returned.

        Returns
        -------
        :
            Random values of specified shape.
        """
        raise NotImplementedError


class UniformSampler(Sampler):
    """Sampler for univariate uniformly distributed samples on given domain.

    Parameters
    ----------
    domain : array_like
        Interval of support of distribution.
    """

    def __init__(self, domain: Union[List, Tuple, np.ndarray]) -> None:
        """Initiate UniformSampler object."""
        self.domain = np.reshape(np.array(domain), (-1, 2))
        assert self.domain.shape == (1, 2)
        self._length = self.domain[0, 1] - self.domain[0, 0]

    @property
    def dimension(self) -> int:
        """Parameter dimension."""
        return self.domain.shape[0]

    @property
    def mass(self) -> float:
        """Mass of the PDF."""
        return 1

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        return 1 / self._length

    @property
    def mean(self) -> float:
        """Mean value of the distribution."""
        return 0.5 * (self.domain[0, 0] + self.domain[0, 1])

    @property
    def cov(self) -> float:
        """(Co)Variance of the distribution."""
        return 1 / 12 * (self.domain[0, 1] - self.domain[0, 0])**2

    @property
    def var(self) -> float:
        """Variance of the distribution."""
        return self.cov

    @property
    def std(self) -> float:
        """Standard deviation of the distribution."""
        return np.sqrt(self.var)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate uniform PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        return scipy.stats.uniform.pdf(
            x, loc=self.domain[0, 0], scale=self._length)

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate uniform log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        return scipy.stats.uniform.logpdf(
            x, loc=self.domain[0, 0], scale=self._length)

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate gradient of uniform log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        return np.zeros_like(x)

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate Hessian of uniform log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        return np.zeros_like(x)

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """ Draw samples from uniform distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        return np.random.uniform(self.domain[0, 0], self.domain[0, 1], shape)


class NormalSampler(Sampler):
    """Sampler for univariate normally distributed samples.

    Parameters
    ----------
    mean : float
        Mean of the Gaussian distribution.
    var : float
        Variance of the Gaussian distribution.
    """

    def __init__(self, mean: float, var: float) -> None:
        """Initiate NormalSampler object."""
        self.domain = np.array([-np.inf, np.inf], ndmin=2)  # shape is (1,2)
        self._mean = mean  # work-around to comply with ABC Sampler class
        assert var >= 0
        self.var = var

    @property
    def mass(self) -> float:
        """Mass of the PDF."""
        return 1

    @property
    def dimension(self) -> float:
        """Dimension of the parameters."""
        return self.domain.shape[0]

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        return 1 / np.sqrt(2*np.pi*self.var)

    @property
    def mean(self) -> float:
        """Mean value of the distribution."""
        return self._mean

    @property
    def cov(self) -> float:
        """(Co)Variance of the distribution."""
        return self.var

    @property
    def std(self) -> float:
        """Standard deviation."""
        return np.sqrt(self.var)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        return scipy.stats.norm.pdf(
            x, loc=self.mean, scale=np.sqrt(self.var))

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        return scipy.stats.norm.logpdf(
            x, loc=self.mean, scale=np.sqrt(self.var))

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate gradient of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        return - (x - self.mean) / self.var

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate Hessian of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        return - 1 / self.var * np.ones_like(x)

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """ Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        return np.random.normal(self.mean, np.sqrt(self.var), shape)


class GammaSampler(Sampler):
    """Sampler for univariate Gamma distributed samples on given domain.

    Parameters
    ----------
    domain : array_like
        Supported domain of distribution.
    alpha : float
        Parameter for Gamma distribution.
    beta : float
        Parameter for Gamma distribution.
    """

    def __init__(self,
                 domain: Union[List, Tuple, np.ndarray],
                 alpha: float,
                 beta: float
                 ) -> None:
        """Initiate GammaSampler object."""
        self.domain = np.reshape(np.array(domain), (-1, 2))
        assert self.domain.shape == (1, 2)
        assert self.domain[0, 1] == np.inf
        self.alpha = alpha
        self.beta = beta
        assert self.alpha > 0 and self.beta > 0

    @property
    def dimension(self) -> float:
        """Dimension of the parameters."""
        return 1

    @property
    def mass(self) -> float:
        """Mass of the PDF."""
        return 1

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF.

        The maximum of the Gamma distribution is given by

        .. math::
            \\max_{x\\in[a,\\infty)} f(x) =
            \\begin{cases}
            \\infty & \\mbox{if } 0 < \\alpha < 1\\\\
            \\frac{\\beta^\\alpha}{\\Gamma(\\alpha)} & \\mbox{if } \\alpha = 1\\\\
            \\frac{\\beta^\\alpha}{\\Gamma(\\alpha)} \\Bigl(\\frac{\\alpha-1}{\\beta} \\Bigr)^{\\alpha-1} e^{1-\\alpha} & \\mbox{if } \\alpha > 1\\\\
            \\end{cases}
        """
        if self.alpha < 1:
            return np.inf

        if self.alpha == 1:
            return self.beta**self.alpha / gamma(self.alpha)

        return (self.beta**self.alpha / gamma(self.alpha)
                * ((self.alpha-1)/self.beta)**(self.alpha-1)
                * np.exp(-(self.alpha-1)))

    @property
    def mean(self) -> float:
        """Mean value of the distribution."""
        return self.alpha / self.beta + self.domain[0, 0]

    @property
    def cov(self) -> float:
        """(Co)Variance of the distribution."""
        return self.alpha / self.beta**2

    @property
    def var(self) -> float:
        """Variance of the distribution."""
        return self.cov

    @property
    def std(self) -> float:
        """Standard deviation of the distribution."""
        return np.sqrt(self.var)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        y = x - self.domain[0, 0]
        return scipy.stats.gamma.pdf(y, a=self.alpha, scale=1.0/self.beta)

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        y = x - self.domain[0, 0]
        return scipy.stats.gamma.logpdf(y, a=self.alpha, scale=1.0/self.beta)

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate gradient of log-PDF.

        .. note::
            Not yet implemented.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of log-PDF.

        .. note::
            Not yet implemented.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        ret = np.random.gamma(self.alpha, 1.0/self.beta, shape)
        return ret + self.domain[0, 0]


class BetaSampler(Sampler):
    """Sampler for univariate Beta distributed samples on given domain.

    Parameters
    ----------
    domain : array_like
        Supported domain of distribution.
    alpha : float
        Parameter for Beta distribution.
    beta : float
        Parameter for Beta distribution.
    """

    def __init__(self,
                 domain: Union[List, Tuple, np.ndarray],
                 alpha: float,
                 beta: float
                 ) -> None:
        """Initiate BetaSampler object."""
        self.domain = np.reshape(np.array(domain), (-1, 2))
        assert self.domain.shape == (1, 2)
        self.length = self.domain[0, 1] - self.domain[0, 0]
        self.alpha = alpha
        self.beta = beta
        assert self.alpha > 0 and self.beta > 0

    @property
    def dimension(self):
        """Dimension of the parameters."""
        return 1

    @property
    def mass(self):
        """Mass of the PDF."""
        return 1

    @property
    def maximum(self):
        """Maximum value of the PDF.

        The maximum of the Beta distribution is given by

        .. math::
            \\max_{x\\in[a,b]} f(x) =
            \\begin{cases}
            \\infty & \\mbox{if } 0 < \\alpha < 1 \\mbox{ or } 0 < \\beta < 1,\\\\
            \\frac{1}{(b-a)B(\\alpha,\\beta)} & \\mbox{if } \\alpha = 1 \\mbox{ or } \\beta = 1,\\\\
            \\frac{(\\alpha-1)^{\\alpha-1}(\\beta-1)^{\\beta-1}}{(\\alpha+\\beta-2)^{\\alpha+\\beta-2}(b-a)B(\\alpha,\\beta)} & \\mbox{if } \\alpha > 1 \\mbox{ and } \\beta > 1, \\\\
            \\end{cases}

        where :math:`B(\\alpha,\\beta)` denotes the Beta-function.
        """
        if self.alpha < 1 or self.beta < 1:
            return np.inf

        Beta = gamma(self.alpha)*gamma(self.beta)/gamma(self.alpha+self.beta)
        if self.alpha == 1 or self.beta == 1:
            return 1 / (Beta * self.length)

        val = ((self.alpha-1)**(self.alpha-1)
               * (self.beta-1)**(self.beta-1)
               / (self.alpha+self.beta-2)**(self.alpha+self.beta-2))
        return val / (Beta * self.length)

    @property
    def mean(self) -> float:
        """Mean value of the distribution."""
        shift = self.domain[0, 0]
        scale = self.length
        return scale * self.alpha / (self.alpha + self.beta) + shift

    @property
    def cov(self) -> float:
        """(Co)Variance of the distribution."""
        a = self.alpha
        b = self.beta
        return a*b / ((a+b)**2 * (a+b+1)) * self.length

    @property
    def var(self) -> float:
        """Variance of the distribution."""
        return self.cov

    @property
    def std(self) -> float:
        """Standard deviation of the distribution."""
        return np.sqrt(self.var)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        y = pt.misc.shiftCoord(x, self.domain.flatten(), [0, 1])
        ret = scipy.stats.beta.pdf(y, a=self.alpha, b=self.beta)
        return ret / self.length

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        y = pt.misc.shiftCoord(x, self.domain.flatten(), [0, 1])
        ret = scipy.stats.beta.logpdf(y, a=self.alpha, b=self.beta)
        return ret - np.log(self.length)

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate gradient of log-PDF.

        .. note::
            Not yet implemented.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of log-PDF.

        .. note::
            Not yet implemented.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        SAMPLE = np.random.beta(self.alpha, self.beta, shape)
        return pt.misc.shiftCoord(SAMPLE, [0, 1], self.domain.flatten())


class WLSUnivariateSampler(Sampler):
    """Sampler for univariate optimally distributed samples on given domain.

    Parameters
    ----------
    domain : array_like
        Interval of support of distribution.
    """

    def __init__(self,
                 param: pt.parameter.Parameter,
                 deg: int,
                 tsa: bool = True
                 ) -> None:
        """Initiate WLSUnivariateSampler object."""
        self.parameter = param
        self.deg = deg
        self._base_sampler = assign_sampler(param)
        self.domain = self._base_sampler.domain
        self._basis = pt.basis.univariate_basis(
            [self.parameter], [self.deg])[0]
        self._tsa = tsa
        self._trial_sampler, self._bulk = self._compute_trial_sampler()
        assert self.domain.shape == (1, 2)
        # self._length = self.domain[0, 1] - self.domain[0, 0]

    @property
    def dimension(self) -> int:
        """Parameter dimension."""
        return self.domain.shape[0]

    @property
    def mass(self) -> float:
        """Mass of the PDF."""
        return 1

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        if self.parameter.distribution == "uniform":
            # alternatively return pdf in domain[0, 1]
            return self.pdf(self.domain[0, 0])
        elif self.parameter.distribution == "normal":
            if self.deg % 2 == 0:
                return self.pdf(self.parameter.mean)
            else:
                # Note: this can be improved
                return get_maximum(self.pdf, self.domain)
        elif self.parameter.distribution == "gamma":
            # Note: this is probably not going to work for domain = [a, inf]
            return get_maximum(self.pdf, self.domain)
        elif self.parameter.distribution == "beta":
            # Note: this can be improved
            return get_maximum(self.pdf, self.domain)
        return get_maximum(self.pdf, self.domain)

    @property
    def mean(self) -> float:
        """Mean value of the distribution."""
        mean, err = quad(
            lambda x: x*self.pdf(x), self.domain[0, 0], self.domain[0, 1])
        if err > 1e-8:
            warnings.warn(f"quadrature error large: ({err})")
        return mean

    @property
    def cov(self) -> float:
        """(Co)Variance of the distribution."""
        moment, err = quad(
            lambda x: x**2*self.pdf(x), self.domain[0, 0], self.domain[0, 1])
        if err > 1e-8:
            warnings.warn(f"quadrature error large: ({err})")
        return moment - self.mean**2

    @property
    def var(self) -> float:
        """Variance of the distribution."""
        return self.cov

    @property
    def std(self) -> float:
        """Standard deviation of the distribution."""
        return np.sqrt(self.var)

    def weight(self, x: Union[np.ndarray, float, int]) -> np.ndarray:
        """Weights for the pdf.

        Parameters
        ----------
        x : np.ndarray
            Points the weight function is evaluated in.

        Returns
        -------
        w : array_like
            Weights of evaluation points `x`.
        """
        if isinstance(x, (float, int)):
            x = np.array(x).reshape(1)
        if x.ndim == 2 and x.shape[1] == 1:
            x = np.array(x).reshape(-1)
        assert x.ndim < 2
        b_eval = np.sum([np.abs(p(x))**2 for p in self._basis], axis=0)
        return (self.deg + 1) / b_eval

    def pdf(self, x: Union[np.ndarray, float, int]) -> np.ndarray:
        """Evaluate uniform PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        if isinstance(x, (float, int)):
            x = np.array(x).reshape(1)
        if x.ndim == 2 and x.shape[1] == 1:
            x = np.array(x).reshape(-1)
        return self._base_sampler.pdf(x) / self.weight(x)

    def log_pdf(self, x: Union[np.ndarray, float, int]) -> np.ndarray:
        """Evaluate uniform log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        if isinstance(x, (float, int)):
            x = np.array(x).reshape(1)
        if x.ndim == 2 and x.shape[1] == 1:
            x = np.array(x).reshape(-1)
        return self._base_sampler.log_pdf(x) - np.log(self.weight(x))

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate gradient of uniform log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of uniform log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def _compute_trial_sampler(self) -> Tuple[Sampler, float]:
        """Trial sampler adaptation.

        .. note::
            TSA currently only available for uniform parameter distribution.

        Parameters
        ----------
        tsa : bool
            Adapt trial sampler or simply use uniform product sampler.

        Returns
        -------
        trial_sampler : pt.sampler.Sampler
            Trial sampler.
        bulk : array_like
            Domain estimate of the mass of the distribution.
        """
        if (self._tsa is True and
                self.parameter.distribution == "uniform"):
            trial_sampler = BetaSampler(self.parameter.domain, 0.5, 0.5)
        else:
            trial_sampler = UniformSampler(self.parameter.domain)
        bulk = get_maximum(
            lambda x: self.pdf(x)/trial_sampler.pdf(x), self.domain
        )
        return trial_sampler, bulk

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from weighted least-squares parameter distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        if isinstance(shape, int):
            shape = (shape,)
        samples = rejection_sampling(
            self.pdf, self._trial_sampler, self._bulk, self.dimension, shape)
        return samples


class ProductSampler(Sampler):
    """Tensor sampler for independent parameters.

    Sampler for cartesian product samples of a list of (independent) univariate
    samplers.

    Parameters
    ----------
    sampler_list : list of `pythia.sampler.Sampler`
        List of (univariate) Sampler objects.
    """

    def __init__(self, sampler_list: List[Sampler]) -> None:
        """ Initiate ProductSampler object. """
        self.samplers = list(sampler_list)
        self.domain = np.squeeze(np.array([s.domain for s in self.samplers]))
        # Add dimension if len(sampler_list) == 1.
        if self.domain.ndim < 2:
            self.domain.shape = 1, 2
        # sampler_list should contain 1D samplers only
        assert self.domain.shape == (len(sampler_list), 2)

    @ property
    def dimension(self) -> int:
        """Dimension of the parameters."""
        return self.domain.shape[0]

    @ property
    def mass(self) -> float:
        """Mass of the PDF."""
        return np.prod(np.array([s.mass for s in self.samplers]))

    @ property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        return np.prod(np.array([s.maximum for s in self.samplers]))

    @ property
    def mean(self) -> np.ndarray:
        """Mean of the PDF."""
        return np.array([s.mean for s in self.samplers])

    @ property
    def cov(self) -> np.ndarray:
        """Covariance of the PDF."""
        return np.diag([s.cov for s in self.samplers])

    def weight(self, x: np.ndarray) -> np.ndarray:
        """Weights of the product PDF.

        Parameters
        ----------
        x : np.ndarray
            Evaluation points.

        Returns
        -------
        :
            Array of uniform weights for samples.
        """
        if x.ndim == 1:
            x.shape = 1, -1
        assert x.ndim == 2
        return np.ones(x.shape[0])

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate PDF.

        The PDF is given by the product of the univariate PDFs.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        assert x.shape[-1] == self.dimension
        densities = [s.pdf for s in self.samplers]
        val = np.array(
            [densities[jj](x[..., jj]) for jj in range(self.dimension)])
        return np.prod(val, axis=0)

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate log-PDF.

        The log-PDF is given by the sum of the univariate log-PDFs.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        assert x.shape[-1] == self.dimension
        densities = [s.log_pdf for s in self.samplers]
        val = np.array(
            [densities[jj](x[..., jj]) for jj in range(self.dimension)])
        return np.sum(val, axis=0)

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate gradient of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        grad_densities = [s.grad_x_log_pdf for s in self.samplers]
        return np.array([grad_densities[jj](x[..., jj])
                         for jj in range(self.dimension)]).T

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        # create an (x.shape[0], self._dim, self._dim) tensor where each
        # (self._dim, self._dim) matrix is the identity
        eye = np.tile(np.expand_dims(np.eye(self.dimension), 0),
                      (x.shape[0], 1, 1))

        # create an (x.shape[0],self._dim,1) tensor where each (x.shape[0],1,1)
        # subtensor is a diagonal entry of the hessian
        hess_densities = [s.hess_x_log_pdf for s in self.samplers]
        hess = np.expand_dims(np.array([hess_densities[jj](x[..., jj])
                                        for jj in range(self.dimension)]).T, 2)
        return eye*hess

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        if isinstance(shape, int):
            shape = (shape,)
        samples = [s.sample(shape) for s in self.samplers]
        return np.stack(samples, -1)


class ParameterSampler(Sampler):
    """Product sampler of given parameters.

    Parameters
    ----------
    params : list of `pythia.parameter.Parameter`
        List containing information of parameters.
    """

    def __init__(self, params: List[pt.parameter.Parameter]) -> None:
        """Initiate ParameterSampler object."""
        self.parameter = params
        assert isinstance(self.parameter, list)
        self.domain = np.array([param.domain for param in self.parameter])
        self._product_sampler = ProductSampler(
            [assign_sampler(param) for param in self.parameter]
        )

    @property
    def dimension(self) -> int:
        """Dimension of the parameters."""
        return self._product_sampler.dimension

    @property
    def mass(self) -> float:
        """Mass of the PDF."""
        return self._product_sampler.mass

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        return self._product_sampler.maximum

    @property
    def mean(self) -> np.ndarray:
        """Mean of the PDF."""
        return self._product_sampler.mean

    @property
    def cov(self) -> np.ndarray:
        """Covariance of the PDF."""
        return self._product_sampler.cov

    def weight(self, x: np.ndarray) -> np.ndarray:
        """Weights of the parameter product PDF.

        Parameters
        ----------
        x : np.ndarray
            Evaluation points.

        Returns
        -------
        :
            Array of uniform weights for samples.
        """
        return self._product_sampler.weight(x)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        return self._product_sampler.pdf(x)

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate log-PDF.

        The log-PDF is given by the sum of the univariate log-PDFs.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        return self._product_sampler.log_pdf(x)

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate gradient of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        return self._product_sampler.grad_x_log_pdf(x)

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        return self._product_sampler.hess_x_log_pdf(x)

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        return self._product_sampler.sample(shape)


class WLSSampler(Sampler):
    """Weighted Least-Squares sampler as in Cohen & Migliorati 2017.

    Parameters
    ----------
    params : list of `pythia.parameter.Parameter`
        List of parameters.
    basis : list
        List of basis functions.
    tsa : bool, default=False
        Trial sampler adaptation. If True, a trial sampler is chosen on the
        distributions of parameters, if false a uniform trial sampler is used.

    Other Parameters
    ----------------
    trial_sampler : pythia.sampler.Sampler, default=None
        Trial sampler for rejection sampling. If `tsa` is true and either
        `trial_sampler` or `bulk` are `None`, the trial sampler is chosen
        automatically.
    bulk : float, defaul=None
        Scaling for trial sampler. If `tsa` is true and either
        `trial_sampler` or `bulk` are `None`, the trial sampler is chosen
        automatically.
    """

    def __init__(self,
                 params: List[pt.parameter.Parameter],
                 basis: List[Callable],
                 tsa: bool = True,
                 trial_sampler: Optional[Sampler] = None,
                 bulk: Optional[float] = None
                 ) -> None:
        """Initiate WLSSampler object."""
        self.parameter = params
        self._param_sampler = ParameterSampler(self.parameter)
        self.domain = self._param_sampler.domain
        self.basis = basis
        self._tsa = tsa
        if trial_sampler is None or bulk is None:
            self._trial_sampler, self._bulk = self._compute_trial_sampler()
        else:
            self._trial_sampler = trial_sampler
            self._bulk = bulk

    @property
    def dimension(self) -> int:
        """Dimension of the parameters."""
        return self._param_sampler.dimension

    @property
    def mass(self):
        """Mass of the PDF."""
        return 1

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        return get_maximum(self.pdf, self.domain)

    @property
    def mean(self) -> np.ndarray:
        """Mean of the PDF."""
        raise NotImplementedError

    @property
    def cov(self) -> np.ndarray:
        """Covariance of the PDF."""
        raise NotImplementedError

    def weight(self, x: np.ndarray) -> np.ndarray:
        """Weights for the PDF.

        Parameters
        ----------
        x : array_like
            Points the weight function is evaluated in.

        Returns
        -------
        :
            weights of evaluation points `x`.
        """
        km = np.sum([abs(p(x))**2 for p in self.basis], axis=0)
        return len(self.basis) / km

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """ Evaluate PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        assert x.shape[-1] == self.dimension
        return self._param_sampler.pdf(x) / self.weight(x)

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate log-PDF.

        The log-PDF is given by the sum of the univariate log-PDFs.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        assert x.shape[-1] == self.dimension
        return self._param_sampler.log_pdf(x) - np.log(self.weight(x))

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate gradient of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        raise NotImplementedError

    def _compute_trial_sampler(self) -> Tuple[Sampler, float]:
        """Trial sampler adaptation.

        Parameters
        ----------
        tsa : bool
            Adapt trial sampler or simply use uniform product sampler.

        Returns
        -------
        trial_sampler : pt.sampler.Sampler
            Trial sampler.
        bulk : array_like
            Domain estimate of the mass of the distribution.
        """
        sampler_list = []
        for param in self.parameter:
            if (self._tsa is True and param.distribution == "uniform"):
                sampler_list.append(BetaSampler(param.domain, 0.5, 0.5))
            else:
                sampler_list.append(UniformSampler(param.domain))
        trial_sampler = ProductSampler(sampler_list)
        bulk = get_maximum(
            lambda x: self.pdf(x)/trial_sampler.pdf(x), self.domain
        )
        return trial_sampler, bulk

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        if isinstance(shape, int):
            shape = (shape,)
        samples = rejection_sampling(
            self.pdf, self._trial_sampler, self._bulk, self.dimension, shape)
        return samples


class WLSTensorSampler(Sampler):
    """ WLS sampler for tensor multivariate basis.

    Sampler for weighted Least-Squares sampling as described by Cohen and
    Migliorati. Only full tensor space can be sampled. This allows for
    univariate weighted Least-Squares sampling in each component.

    Parameters
    ----------
    params : list of `pythia.parameter.Parameter`
        Parameter list.
    deg : list of int
        Polynomial degree of each component (same for all).
    tsa : bool, default=True
        Trial sampler adaptation. If True, a trial sampler is chosen on the
        distributions of parameters, if false a uniform trial sampler is
        used.
    """

    def __init__(self,
                 params: List[pt.parameter.Parameter],
                 deg: List[int],
                 tsa: bool = True
                 ) -> None:
        """ Initiate WLSTensorSampler object. """
        self.parameter = params
        self.deg = deg
        self.domain = np.array([param.domain for param in self.parameter])
        self._tsa = tsa
        # self._uBasis = pt.basis.univariate_basis(self.parameter, self.deg)
        # self._univariate_samplers = [assign_sampler(param)
        #                              for param in self.parameter]
        self._univariate_samplers = [WLSUnivariateSampler(
            p, d, self._tsa) for (p, d) in zip(self.parameter, self.deg)]
        self._product_sampler = ProductSampler(self._univariate_samplers)

    @property
    def dimension(self) -> int:
        """Dimension of the parameters."""
        return self._product_sampler.dimension

    @property
    def mass(self) -> float:
        """Mass of the PDF."""
        return self._product_sampler.mass

    @property
    def maximum(self) -> float:
        """Maximum value of the PDF."""
        return self._product_sampler.maximum

    @property
    def mean(self) -> np.ndarray:
        """Mean of the PDF."""
        return self._product_sampler.mean

    @property
    def cov(self) -> np.ndarray:
        """Covariance of the PDF."""
        return self._product_sampler.cov

    def weight(self, x: np.ndarray) -> np.ndarray:
        """Weights for the PDF.

        Parameters
        ----------
        x : array_like
            Points the weight function is evaluated in.

        Returns
        -------
        :
            Weights of evaluation points `x`.
        """
        assert x.shape[-1] == self.dimension
        weights = [s.weight for s in self._univariate_samplers]
        val = np.array(
            [weights[jj](x[..., jj]) for jj in range(self.dimension)])
        return np.prod(val, axis=0)

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of PDF evaluated in `x`.
        """
        return self._product_sampler.pdf(x)

    def log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate log-PDF.

        The log-PDF is given by the sum of the univariate log-PDFs.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of log-PDF evaluated in `x`.
        """
        return self._product_sampler.log_pdf(x)

    def grad_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate gradient of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of gradient (vector valued) of log-PDF evaluated in `x`.
        """
        return self._product_sampler.grad_x_log_pdf(x)

    def hess_x_log_pdf(self, x: np.ndarray) -> np.ndarray:
        """Evaluate Hessian of log-PDF.

        Parameters
        ----------
        x : array_like
            Evaluation points.

        Returns
        -------
        :
            Values of Hessian (matrix valued) of log-PDF evaluated in `x`.
        """
        return self._product_sampler.hess_x_log_pdf(x)

    def sample(self, shape: Union[int, List, Tuple, np.ndarray]) -> np.ndarray:
        """Draw samples from distribution.

        Parameters
        ----------
        shape : array_like
            Shape of the samples.

        Returns
        -------
        :
            Random samples of specified shape.
        """
        return self._product_sampler.sample(shape)


def rejection_sampling(pdf: Callable,
                       trial_sampler: Sampler,
                       scale: float,
                       dimension: int,
                       shape: Union[int, List, Tuple, np.ndarray],
                       ) -> np.ndarray:
    """Draw samples from pdf by rejection sampling.

    Parameters
    ----------
    pdf : Callable
        Probability density the samples are generated from.
    trial_sampler : Sampler
        Trial sampler proposal samples are drawn from.
    scale : float
        Threshold parameter with ``pdf <= scale * trialSampler.pdf``
    dimension : int
        Dimension of the (input of the) pdf.
    shape : array_like
        Shape of the samples.

    Returns
    -------
    :
        Random samples of specified shape.
    """
    if isinstance(shape, int):
        shape = (shape,)
    size = np.prod(np.array(shape))
    trial_samples = trial_sampler.sample(size)
    samples = np.empty_like(trial_samples)
    pointer = 0
    while pointer < size:
        trial_samples = trial_sampler.sample(size)
        checker = np.random.rand(trial_samples.shape[0])
        is_valid = (
            checker * max(scale, 1) * trial_sampler.pdf(trial_samples)
            <= pdf(trial_samples)
        )
        valid_samples = trial_samples[is_valid]
        start = pointer
        end = min(pointer+valid_samples.shape[0], size)
        samples[start:end] = valid_samples[0:min(
            valid_samples.shape[0], size-pointer)]
        pointer = end
    if dimension > 1:
        return np.moveaxis(samples.T.reshape(dimension, *shape), 0, -1)
    return samples.reshape(*shape)


def constraint_sampling(sampler: Sampler,
                        constraints: List[Callable],
                        shape: Union[int, List, Tuple, np.ndarray],
                        ) -> np.ndarray:
    """Draw samples according to algebraic constraints.

    Draw samples from target distribution and discard samples that do not
    satisfy the constraints.

    Parameters
    ----------
    sampler : Sampler
        Sampler to sample from.
    constraints : list of callable
        List of functions that return True if sample point satisfies the
        constraint.

    Returns
    -------
    :
        Samples drawn from sampler satisfying the constraints.

    Notes
    -----
    The constaints may lead to a non-normalized density function.
    """
    if isinstance(shape, int):
        shape = (shape,)
    samples = np.empty(sampler.sample(np.prod(shape)).shape)
    for j_sample in range(samples.shape[0]):
        proposal = sampler.sample((1,))
        met_constraints = np.all([c(proposal) for c in constraints])
        while not met_constraints:
            proposal = sampler.sample((1,))
            met_constraints = np.all([c(proposal) for c in constraints])
        samples[j_sample] = proposal
    return np.moveaxis(samples.T.reshape(sampler.dimension, *shape), 0, -1)


def assign_sampler(param: pt.parameter.Parameter) -> Sampler:
    """Assign a univariate sampler to the given parameter.

    Parameters
    ----------
    param : pythia.parameter.Parameter

    Returns
    -------
    :
        Univariate sampler.
    """
    if param.distribution == "uniform":
        return UniformSampler(param.domain)
    elif param.distribution == "normal":
        return NormalSampler(param.mean, param.var)
    elif param.distribution == "gamma":
        return GammaSampler(param.domain, param.alpha, param.beta)
    elif param.distribution == "beta":
        return BetaSampler(param.domain, param.alpha, param.beta)
    raise ValueError(f"unknown distribution '{param.distribution}'")


def get_maximum(f: Callable,
                domain: Union[List, Tuple, np.ndarray],
                n: int = 1000
                ) -> float:
    """Compute essential maximum of function by point evaluations.

    Parameters
    ----------
    f : callable
        Function to evaluate. Needs to map from n-dim space to 1-dim space.
    domain : array_like
        Domain to evaluate function on.
    n : int, default=1000
        Number of function evaluations. Evaluations are done on a uniform grid
        in domain. Actual number of points may thus be a little greater.

    Returns
    -------
    :
        Approximation of maximum of function `f`.
    """
    if not isinstance(domain, np.ndarray):
        domain = np.array(domain)
    assert domain.shape[-1] == 2 and domain.ndim == 2
    n_points = int(np.ceil(np.power(n, 1/domain.shape[0])))
    if domain.shape[0] > 1:
        eps = np.finfo(float).eps  # circumvent inf on bdry
        x = pt.misc.cartProd([
            np.linspace(dom[0]+eps, dom[1]-eps, n_points) for dom in domain]
        )
    else:
        x = np.linspace(domain[0, 0], domain[0, 1], n_points).reshape(-1, 1)
    return np.max(np.abs(f(x)))


if __name__ == "__main__":
    pass



""" Sample-based computation of polynomial chaos expansion.

This module provides a class to compute the polynomial chaos approximation of
an unknown function given via input/output training data pairs via linear
least-squares regression.
"""
from typing import List, Tuple, Optional, Union
import psutil
import warnings
import numpy as np
import pythia as pt


class PolynomialChaos:
    """Computation of sparse polynomial chaos expansion.

    Parameters
    ----------
    params : list of `pt.parameter.Parameter`
        List of stochastic parameters.
    index_set : pt.index.IndexSet
        Index set for sparse polynomial chaos expansion.
    x_train : array_like
        Parameter realizations for training.
    weights : array_like
        Regression weights for training.
    fEval : array_like
        Function evaluation for training.
    coefficients : array_like, default=None
        Polynomial expansion coefficients. If given, the coefficients are not
        computed during initiation. This can be used to load a chaos expansion.
    """

    def __init__(self,
                 params: List[pt.parameter.Parameter],
                 index_set: pt.index.IndexSet,
                 x_train: np.ndarray,
                 w_train: np.ndarray,
                 y_train: np.ndarray,
                 coefficients: Optional[np.ndarray] = None
                 ) -> None:
        """Initiate the computation of the PC expansion of a function."""
        assert x_train.ndim == 2 and x_train.shape[1] == len(params)
        assert w_train.ndim == 1 and w_train.size == x_train.shape[0]
        assert y_train.ndim == 2 and y_train.shape[0] == x_train.shape[0]

        self.parameters = params
        self.index_set = index_set
        self.x_train = x_train
        self.w_train = w_train
        self.y_train = y_train

        self.n_samples, self.sdim = x_train.shape
        self.ydim = y_train.shape[1]

        self.univariate_pdfs = [pt.sampler.assign_sampler(param)
                                for param in self.parameters]
        self.pdf = pt.sampler.ParameterSampler(self.parameters).pdf
        self.univariate_bases = pt.basis.univariate_basis(
            self.parameters, self.index_set.max)
        self.basis = pt.basis.multivariate_basis(
            self.univariate_bases, self.index_set.indices)

        self.gramian, self.basis_eval_mat = self._assemble_matrices()
        if coefficients is None:
            self.coefficients = self._compute_pc_coefficients()
        else:
            assert coefficients.shape == (index_set.shape[0], self.ydim)
            self.coefficients = coefficients
        self.sobol = self._compute_sobol_indices()

    @property
    def mean(self) -> np.ndarray:
        """Mean of the PC expansion."""
        idx = self.index_set.get_index_number(
            np.zeros((1, self.sdim), dtype=int))
        return self.coefficients[idx].flatten()

    @property
    def var(self) -> np.ndarray:
        """Variance of the PC expansion."""
        if (self.index_set.shape[0] == 1 and
                np.all(self.index_set.indices == 0)):
            return np.zeros(self.ydim)
        return np.sum(self.coefficients**2, axis=0).flatten() - self.mean**2

    @property
    def std(self) -> np.ndarray:
        """Standard deviation of the PC expansion."""
        return np.sqrt(self.var)

    def _assemble_matrices(self) -> Tuple[np.ndarray, np.ndarray]:
        """Assemble Gramian and basis evaluation matrix.

        Assemble the information matrix :math:`A` and the basis evaluation
        matrix :math:`\\Psi` with the regression points of the PC expansion.
        The basis evaluation matrix :math:`\\Psi` is given by

        .. math::
            \\Psi_{kj} = \\operatorname{basis}[j](\\operatorname{regPoints}[k]).

        Returns
        -------
        gramian : np.ndarray
            Empirical Gramian matrix.
        psi_mat : np.ndarray
            Basis evaluation matrix :math:`\\Psi`.
        """
        batch_size = get_gram_batchsize(len(self.basis))
        psi_mat = np.array([p(self.x_train) for p in self.basis])
        gramian = np.zeros((len(self.basis), len(self.basis)))
        for batch in pt.misc.batch(range(self.x_train.shape[0]), batch_size):
            mat_1 = psi_mat[:, batch].reshape(1, len(self.basis), len(batch))
            mat_2 = psi_mat[:, batch].reshape(len(self.basis), 1, len(batch))
            gramian += np.sum(self.w_train[batch] * mat_1 * mat_2, axis=-1)
        return gramian, psi_mat.T

    def _compute_pc_coefficients(self) -> np.ndarray:
        """Compute polynomial chaos expansion coefficients.

        Compute the PC coefficients with linear regression. The coefficients
        are given by

        .. math::
            S = A^(-1) * \\Psi^T * W * F_\\mathrm{ex}

        where the Gram matrix :math:`A` is of full rank but may be ill
        conditioned.  :math:`F_\\mathrm{ex}` is an array containing the values
        of f evaluated at the required regression points. For more detail on
        the Gram matrix or :math:`\\Psi`, see `assemble_matrices()`.

        Returns
        -------
        :
            Polynomial chaos expansion coefficients.
        """
        u, s, vh = np.linalg.svd(self.gramian)
        gramian_inv = np.dot(vh.T, np.dot(np.diag(1/s), u.T))
        W = self.w_train.reshape(-1, 1)
        coefficients = np.linalg.multi_dot(
            [gramian_inv, self.basis_eval_mat.T, W*self.y_train])
        return coefficients

    def _compute_sobol_indices(self) -> np.ndarray:
        """Compute Sobol indices.

        The Sobol coefficients are given as

        .. math::
            S_{i_1,...,i_k} = \\sum_{\\alpha\\in\\mathcal{M}} f_\\alpha(x)^2

        where :math:`\\mathcal{M} = { \\alpha | \\alpha_{i_j} != 0 for j = 1,...,k }`.
        """
        sobol = np.zeros([len(self.index_set.sobol_tuples), self.ydim])
        # mask components of f with zero variance
        nz_idx = np.nonzero(self.var)
        for idx, sdx in enumerate(self.index_set.sobol_tuples):
            _mdx = self.index_set.sobol_tuple_to_indices([sdx])[0]
            rows = self.index_set.get_index_number(_mdx)
            coeffs = self.coefficients[rows]
            sobol[idx, nz_idx] = (
                np.sum(coeffs[:, nz_idx]**2, axis=0) / self.var[nz_idx])
        return sobol

    def eval(self,
             x: np.ndarray,
             partial: Optional[List[int]] = None
             ) -> np.ndarray:
        """Evaluate the (partial derivative of the) PC approximation.

        Parameters
        ----------
        x : array_like
            Parameter realizations in which the approximation is evaluated.
        partial : list of int
            List that specifies the number of derivatives in each component.
            Length is the number of parameters.

        Returns
        -------
        :
            Evaluation of polynomial expansion in x values.
        """
        if x.ndim < 1 or x.ndim > 2:
            raise ValueError(f"Wrong ndim: '{x.ndim}'")
        if x.ndim == 1:
            x.shape = 1, -1
        c = self.coefficients.reshape(self.coefficients.shape[0], 1, -1)
        if partial is None:
            basis = self.basis
        else:
            basis = pt.basis.multivariate_basis(
                self.univariate_bases, self.index_set.indices, partial)
        eval_mat = np.array([b(x) for b in basis])
        eval_mat.shape = eval_mat.shape[0], eval_mat.shape[1], 1
        return np.sum(c*eval_mat, axis=0)


def find_optimal_indices(params: List[pt.parameter.Parameter],
                         x_train: np.ndarray,
                         w_train: np.ndarray,
                         y_train: np.ndarray,
                         max_terms: int = 0,
                         threshold: float = 1e-03
                         ) -> Tuple[np.ndarray, np.ndarray]:
    """Compute optimal multiindices of polynomial chaos expansion.

    Compute the optimal multiindices for a polynomial chaos expansion based on
    an estimate of the Sobol coefficient values.

    Parameters
    ----------
    params : list of pythia.Parameters.Parameter
        Random parameters of the problem.
    x_train : array_like
        Sample points for training
    w_train : array_like
        Weights for training.
    y_train : array_like
        Function evaluations for training.
    max_terms : int, default=0
        Maximum number of expansion terms. Number of expansion terms is chosen
        automatically for `max_terms=0`.
    threshold : float, default=1e-03
        Truncation threshold for Sobol indices. Sobol indices with smaller
        magnitude are ignored.

    Returns
    -------
    indices :
        Array with multiindices.
    sobol :
        Crude intermediate approximation of Sobol indices.
    """
    assert 0 <= threshold < 1
    # set maximal number of expansion terms
    n_samples, dim = x_train.shape
    if max_terms > 0:
        _max_terms = max_terms
    else:
        _max_terms = int(n_samples/np.log(n_samples)/2)
    if _max_terms > int(n_samples/np.log(n_samples)/2):
        warnings.warn("Gramian may become ill conditioned.")

    # compute crude approximation of Sobol coefficients
    def fac(n): return float(np.math.factorial(n))
    deg = 1
    for _deg in range(2, 1000):
        n_terms = fac(_deg+dim) / fac(_deg) / fac(dim)
        if n_terms > _max_terms:
            break
        deg = _deg
    _indices = pt.index.simplex(dim, deg)
    index_set = pt.index.IndexSet(_indices)
    surrogate = PolynomialChaos(params, index_set, x_train, w_train, y_train)

    # sort Sobol coefficients by largest and mark top threshold percent.
    idx, vals, marker = pt.misc.doerfler_marking(
        np.sum(surrogate.sobol, axis=1), threshold=1-threshold)

    # assemble adaptive choice of multiindices
    indices = assemble_indices(
        idx[:marker], index_set.sobol_tuples, _max_terms)

    return indices, surrogate.sobol


def assemble_indices(enum_idx: Union[List[int], Tuple[int], np.ndarray],
                     sobol_tuples: List[Tuple[int]],
                     max_terms: int
                     ) -> np.ndarray:
    """Compute automatic choice of multiindices.

    Parameters
    ----------
    enum_idx : np.ndarray
        Sorted enumeration indices according to magnitude of Sobol indices.
    sobol_tuples : list of tuple
        List of Sobol subscript tuples.
    max_terms : int
        Maximum number of expansion terms.

    Returns
    -------
    indices : np.ndarray
        Array of (sparse) optimal multiindices.
    """
    dim = max(len(sdx) for sdx in sobol_tuples)
    n_terms_per_idx = int((max_terms-1)/len(enum_idx))
    indices_list = []
    for idx in enum_idx:
        components = [s-1 for s in sobol_tuples[idx]]  # current sdx
        deg = int(n_terms_per_idx**(1/len(components))+1)
        tmp_indices = pt.index.tensor(
            [deg+2 if j in components else 1 for j in range(dim)],
            [1 if j in components else 0 for j in range(dim)])
        indices_list += [tmp_indices[:n_terms_per_idx]]
    indices = pt.index.sort_index_array(np.concatenate(
        [np.zeros((1, dim))]+indices_list, axis=0))
    return indices


def get_gram_batchsize(dim, save_memory=1025**3/2):
    """Compute memory allocation batch sizes for information matrix.

    Compute the maximal number of samples in each batch when assembling the
    information matrix to be maximally memory efficient and avoid OutOfMemory
    errors.

    Parameters
    ----------
    dim : int
        Number of rows/columns of information matrix.
    save_memory : int, default=3*1025/2
        Memory (in bytes), that should be kept free. The default is equivalent
        to 512 MB.

    Returns
    -------
    n : int
        Batchsize for assembling of information matrix.
    """
    available_memory = psutil.virtual_memory().available
    mem = available_memory - save_memory
    n = int(mem / 8 / dim**2)
    if n < 1:
        # There is less memory available than required for at least one sample.
        raise MemoryError('Not enough free memory.')
    else:
        return n


if __name__ == '__main__':
    pass



"""Create, manipulate and store information about multiindices."""
from typing import List, Tuple, Union, Optional
import itertools
import numpy as np
import pythia as pt


class IndexSet:
    """Generate index set object for sparse PC expansion.

    Parameters
    ----------
    indices : np.ndarray
        Array of multiindices with shape (#indices, param dim).
    """

    def __init__(self, indices: np.ndarray) -> None:
        """Initialize sparse multiindex object."""
        assert indices.ndim == 2 and indices.shape[0] > 0
        assert indices.dtype == int
        assert np.all(indices >= 0)
        self.indices = sort_index_array(indices)
        self.shape = self.indices.shape
        self.max = np.array(np.max(self.indices, axis=0), dtype=int)
        self.min = np.array(np.min(self.indices, axis=0), dtype=int)
        self.sobol_tuples = self._get_sobol_tuple_list()

    def _get_sobol_tuple_list(self) -> List:
        """Generate list of all possible Sobol index id tuples (subscripts).

        Returns
        -------
        :
            List of Sobol tuples.
        """
        sobol_tuples = []
        for r in range(1, self.shape[1]+1):
            sobol_tuples += list(
                itertools.combinations(range(1, self.shape[1]+1), r))
        return sobol_tuples

    def get_index_number(self, indices: np.ndarray) -> np.ndarray:
        """Get enumeration number of indices.

        Get the row indices of the given multiindices such that
        `self.indices[rows] = indices`.

        Parameters
        ----------
        indices : np.ndarray
            Indices to get the number of.

        Returns
        -------
        :
            Array containing the enumeration numbers of the indices.
        """
        return np.array([np.where((self.indices == index).all(axis=1))[0]
                         for index in indices], dtype=int).flatten()

    def get_sobol_tuple_number(self, sobol_tuples: List[tuple]) -> np.ndarray:
        """Get enumeration indices of Sobol tuples.

        Parameters
        ----------
        sobol_tuples : list of tuple
            List of Sobol tuples.

        Returns
        -------
        :
            Array containing the enumeration number of the Sobol tuples.
        """
        return np.array([self.sobol_tuples.index(s) for s in sobol_tuples],
                        dtype=int)

    def index_to_sobol_tuple(self, indices: np.ndarray) -> List[Tuple]:
        """Map array of indices to their respective Sobol tuples.

        Parameters
        ----------
        indices : np.ndarray
            Array of multiindices.

        Returns
        -------
        :
            List of Sobol tuples.
        """
        sobol_tuples = [tuple(np.flatnonzero(index)+1) for index in indices]
        return sobol_tuples

    def sobol_tuple_to_indices(self,
                               sobol_tuples: Union[tuple, List[tuple]]
                               ) -> List[np.ndarray]:
        """Map Sobol tuples to their respective indices.

        Parameters
        ----------
        sobol_tuples : tuple or list of tuple
            List of Sobol tuples.

        Returns
        -------
        :
            List of index arrays for each given Sobol tuple.
        """
        if isinstance(sobol_tuples, tuple):
            sobol_tuples = [sobol_tuples]
        assert isinstance(sobol_tuples, list)
        ret = []
        lookup_dict = {sobol_tuple: [] for sobol_tuple in self.sobol_tuples}
        index_sobol_tuple_list = self.index_to_sobol_tuple(self.indices)
        for sobol_tuple, index in zip(index_sobol_tuple_list, self.indices):
            if len(sobol_tuple) > 0:
                lookup_dict[sobol_tuple] += [index]
        for sobol_tuple in sobol_tuples:
            ret += [np.array(lookup_dict[sobol_tuple], dtype=int)]
        return ret


def sort_index_array(indices: np.ndarray) -> np.ndarray:
    """Sort multiindices and remove duplicates.

    Sort rows of `indices` by sum of multiindex and remove duplicate
    multiindices.

    Parameters
    ----------
    indices : np.ndarray
        Index list before sorting.

    Returns
    -------
    :
        Sorted index array.
    """
    sorted_indices = np.unique(indices, axis=0)
    idx = np.argsort(np.sum(sorted_indices, axis=1))
    sorted_indices = sorted_indices[idx]
    return np.array(sorted_indices, dtype=int)


def add_indices(index_list: List[np.ndarray]) -> np.ndarray:
    """Add multiple arrays of multiindices.

    Concatenate multiple arrays of multiindices, remove duplicates and sort
    them by sum of multiindices.

    Parameters
    ----------
    index_list : list of np.ndarray
        List of multiindex arrays.

    Returns
    -------
    :
        Array with all multiindices.
    """
    all_indices = np.concatenate(index_list, axis=0)
    return sort_index_array(all_indices)


def subtract_indices(indices: np.ndarray, subtract: np.ndarray) -> np.ndarray:
    """Set difference of two index arrays.

    Parameters
    ----------
    indices : np.ndarray
        Index array multiindices are taken out of.
    subtract : np.ndarray
        Indices that are taken out of the original set.

    Returns
    -------
    :
        Set difference of both index arrays.
    """
    indices = sort_index_array(indices)
    subtract = sort_index_array(subtract)
    assert indices.shape[1] == subtract.shape[1]
    idxs = []
    for mdx in subtract:
        idx = np.where((indices == mdx).all(axis=1))[0]
        assert idx.size < 2
        if idx.size == 1:
            idxs += [idx]
    return np.delete(indices, np.array(idxs, dtype=int).flatten(), axis=0)


def tensor(shape: Union[List, Tuple, np.ndarray],
           lower: Optional[Union[List, Tuple, np.ndarray]] = None
           ) -> np.ndarray:
    """Create a tensor index set.

    Parameters
    ----------
    shape : array_like
        Shape of the tensor, enumeration starting from 0.
    lower : array_like, default = None
        Starting values for each dimension of the tensor set. If None, all
        dimensions start with 0.

    Returns
    -------
    :
        Array with all possible multiindices in tensor set.

    Examples
    --------
    >>> pt.index.tensor([2, 2])
    array([[0, 0],
           [0, 1],
           [1, 0],
           [1, 1]])

    It is also possible to use the tensor functions to create n-variate subsets
    for the overall tensor set.

    >>> pt.index.tensor([1, 5, 2], [0, 2, 0])
    array([[0, 2, 0],
           [0, 2, 1],
           [0, 3, 0],
           [0, 3, 1],
           [0, 4, 0],
           [0, 4, 1]])
    """
    shape = np.array(shape, dtype=int)
    if lower is None:
        lower = np.zeros(shape.size, dtype=int)
    elif isinstance(lower, (list, tuple)):
        lower = np.array(lower, dtype=int)
    assert shape.size == lower.size
    univariate_dims = [np.arange(low, up, dtype=int)
                       for low, up in zip(lower, shape)]
    if shape.size > 1:
        ret = pt.misc.cartProd(univariate_dims)
    else:
        ret = np.array(univariate_dims).T
    return sort_index_array(ret)


def simplex(dimension: int, maximum: int) -> np.ndarray:
    """Create a simplex index set.

    A simplex index set consists of all multiindices with sum less then or
    equal the maximum given.

    Parameters
    ----------
    dimension : int
        Dimension of the multiindices.
    maximum : int
        Maximal sum value for the multiindices.

    Returns
    -------
    :
        Array with all possible multiindices in simplex set.

    Examples
    --------
    >>> pt.index.simplex(2, 2)
    array([[0, 0],
           [0, 1],
           [1, 0],
           [0, 2],
           [1, 1],
           [2, 0]])
    """
    assert dimension > 0 and maximum > 0
    all_indices = tensor([maximum+1]*dimension)
    rows = np.where(np.sum(all_indices, axis=1) <= maximum)[0]
    return sort_index_array(all_indices[rows])



"""Miscellaneous functions to support PyThia core functionality."""
import os
import datetime
import shutil
import numpy as np
from typing import List, Tuple, Sequence, Iterator, Optional, Union, Dict


def distributionDict() -> Dict:
    """Set aliases for distribution descriptions.

    .. deprecated:: 2.0.0
        `distributionDict` will be removed in PyThia 3.0.0.

    Returns
    -------
    :
        Dictionary with aliases for distribution descriptions.
    """
    distDict = {}
    distDict['uniform'] = ['uniform', 'Uniform', 'U']
    distDict['normal'] = ['normal', 'gauss', 'Gauss', 'gaussian', 'N']
    distDict['gamma'] = ['gamma', 'Gamma', 'G']
    distDict['beta'] = ['beta', 'Beta', 'B']
    return distDict


def shiftCoord(x: Union[float, np.ndarray],
               T: Union[np.ndarray, List],
               I: Union[np.ndarray, List]
               ) -> np.ndarray:
    """Shift `x` in interval `T` to interval `I`.

    Use an affine transformation to shift points :math:`x` from the interval
    :math:`T = [t_0, t_1]` to the interval :math:`I = [a, b]`.

    Parameters
    ----------
    x : array_like
        Points in interval :math:`T`.
    T : array_like
        Original interval.
    I : array_like
        Target interval.

    Returns
    -------
    :
        Shifted values for `x`.
    """
    return ((I[1]-I[0])*x + I[0]*T[1] - I[1]*T[0]) / (T[1]-T[0])


def cartProd(array_list: Union[List[np.ndarray], np.ndarray]) -> np.ndarray:
    """Compute the outer product of two or more arrays.

    Assemble an array containing all possible combinations of the elements
    of the input vectors :math:`v_1,\\dots,v_n`.

    Parameters
    ----------
    array_list : list of array_like
        List of vectors :math:`v_1,\\dots,v_n`.

    Returns
    -------
    :
        Cartesian product array.
    """
    dim = len(array_list)
    if dim == 1:
        return np.array(array_list).T
    x = np.hstack((np.meshgrid(*array_list))).swapaxes(0, 1).reshape(dim, -1).T
    return x


def is_contained(val: Union[float, Sequence, np.ndarray],
                 domain: Union[List, np.ndarray]
                 ) -> bool:
    """Check if a given value (vector) is contained in a domain.

    Checks if each component of the vector lies in the one dimensional
    interval of the corresponding component of the domain.

    Parameters
    ----------
    val : array_like
        Vector to check containment in domain
    domain : array_like
        Product domain of one dimensional intervals.

    Returns
    -------
    :
        Bool stating if value is contained in domain.
    """
    if not isinstance(val, np.ndarray):
        val = np.array(val)
    if not isinstance(domain, np.ndarray):
        domain = np.array(domain)
    if val.ndim < 2:
        val.shape = 1, -1
    if domain.ndim < 2:
        domain.shape = 1, -1
    assert val.ndim == 2
    assert val.shape[0] == 1
    assert domain.ndim == 2 and domain.shape[1] == 2
    if np.all(domain[:, 0] <= val) and np.all(val <= domain[:, 1]):
        return True
    return False


def formatTime(dt: float) -> str:
    """Converts time (seconds) to time format string.

    Parameters
    ----------
    dt : float
        Time in seconds.

    Returns
    -------
    :
        Formatted time string.
    """
    assert dt >= 0
    dct = {}
    dct['d'], rem = divmod(int(dt), 86400)
    dct['h'], rem = divmod(int(rem), 3600)
    dct['min'], seconds = divmod(int(rem), 60)
    dct['sec'] = seconds+1  # rounding seconds up
    fmt = ''
    if dct['d'] != 0:
        fmt += '{d} days '
    if dct['h'] != 0:
        fmt += '{h}h '
    if dct['min'] != 0:
        fmt += '{min}min '
    if dct['sec'] > 1:
        fmt += '{sec}s '
    if dt < 1.0:
        fmt += '{:2.2g}s'.format(dt)
    fmt = fmt.strip()
    return fmt.format(**dct)


def now() -> str:
    """Get string of current machine date and time.

    Returns
    -------
    :
        Formatted date and time string.
    """
    dt = datetime.datetime.now()
    today = '{:04}-{:02}-{:02} '.format(dt.year, dt.month, dt.day)
    now = '{:02}:{:02}:{:02}'.format(dt.hour, dt.minute, dt.second)
    return today + now


def line(indicator: str, message: str = None) -> str:
    """Print a line of 80 characters by repeating indicator.

    An additional message can be given.

    Parameters
    ----------
    indicator : string
        Indicator the line consists of, e.g. '-', '+' or '+-'.
    message : string, optional
        Message integrated in the line.

    Returns
    -------
    :
        String of 80 characters length.
    """
    assert len(indicator) > 0
    text = ''
    if message is not None:
        text = 2*indicator
        text = text[:2] + ' ' + message + ' '
    while len(text) < 80:
        text += indicator
    return text[:80]


def save(filename: str, data: np.ndarray, path: str = './') -> None:
    """Wrapper for numpy save.

    Assures path directory is created if necessary and backup old data if
    existent.

    Parameters
    ----------
    name : str
        Filename to save data to.
    data : array_like
        Data to save as .npy file.
    path : str, default='./'
        Path under which the file should be created.
    """
    if not os.path.isdir(path):
        os.makedirs(path)
    if os.path.isfile(path+filename):
        shutil.copyfile(path+filename, path+filename+'.backup')
    np.save(path+filename, data)


def load(filename: str) -> np.ndarray:
    """Alias for numpy.load()."""
    return np.load(filename)


def str2iter(string: str,
             iterType: type = list,
             dataType: type = int
             ) -> Sequence:
    """Cast `str(iterable)` to `iterType` of `dataType`.

    Cast a string of lists, tuples, etc to the specified iterable and data
    type, i.e., for `iterType=tuple` and `dataType=float` cast
    ``str([1,2,3]) -> (1.0, 2.0, 3.0)``.

    Parameters
    ----------
    string : str
        String representation of iterable.
    iterType : iterable, default=list
        Iterable type the string is converted to.
    dataType : type, default=int
        Data type of entries of iterable, e.g. `int` or `float`.
    """
    items = [s.strip() for s in string[1:-1].split(",")]
    if items[-1] == "":
        items = items[:-1]
    return iterType([dataType(item) for item in items])


def batch(iterable: Sequence, n: int = 1) -> Iterator:
    """Split iterable into different batches of batchsize n.

    Parameters
    ----------
    iterable : array_like
        Iterable to split.
    n : int, default=1
        Batch size.

    Yields
    ------
    :
        Iterable for different batches.
    """
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def wlsSamplingBound(m: int, c: float = 4) -> int:
    """Compute the weighted Least-Squares sampling bound.

    The number of samples :math:`n` is chosen such that

    .. math::
        \\frac{n}{\\log(n)} \\geq cm,

    where :math:`m` is the dimension of the Gramian matrix (number of PC
    expansion terms) and :math:`c` is an arbitrary constant. In
    Cohen & Migliorati 2017 the authors observed that the coice :math:`c=4`
    yields a well conditioned Gramian with high probability.

    Parameters
    ----------
    m : int
        Dimension of Gramian matrix.
    c : float, default=4
        Scaling constant.

    Returns
    -------
    :
        Number of required wLS samples.
    """
    assert m > 0 and c > 0
    jj = max(int(np.ceil(c*m*np.log(c*m))), 2)
    while True:
        if jj / np.log(jj) >= c*m:
            n = jj
            break
        jj += 1
    return n


def gelman_rubin_condition(chains: np.ndarray) -> np.ndarray:
    """Compute Gelman-Rubin criterion.

    Implementation of the Gelman-Rubin convergence criterion for multiple
    parameters.  A Markov chain is said to be in its convergence, if the final
    ration is close to one.

    Parameters
    ----------
    chains : array_like, ndim=3
        Array containing the Markov chains of each parameter. All chains are
        equal in length, the assumed shape is
        ``(#chains, chain length, #params)``.

    Returns
    -------
    :
        Values computed by Gelman-Rubin criterion for each parameter.
    """
    assert chains.ndim == 3
    M, N, DIM = chains.shape  # chains shape is (#chains, len(chains), #params)

    # Mean and var of chains.
    chain_means = np.mean(chains, axis=1)  # shape is (#chains, #params)
    chain_vars = np.var(chains, axis=1)  # shape is (#chains, #params)

    # Mean across all chains.
    mean = np.mean(chain_means, axis=0).reshape(1, -1)  # shape = (1, #params)

    # Between chain variance.
    B = N / (M-1) * np.sum((chain_means-mean)**2, axis=0)
    # Within chain variance.
    W = 1 / M * np.sum(chain_vars, axis=0)
    # pooled variance
    V = (N-1)/N*W + (M+1)/(M*N)*B

    return np.array([np.sqrt(v/w) if w > 0 else np.inf for v, w in zip(V, W)])


def get_confidence_interval(samples: np.ndarray,
                            rate: float = 0.95,
                            resolution: int = 500
                            ) -> np.ndarray:
    """Compute confidence intervals of samples.

    Compute the confidence intervals of the 1D marginals of the samples
    (slices).  The confidence interval of a given rate is the interval around
    the median (not mean) of the samples containing roughly `rate` percent of
    the total mass.  This is computed for the left and right side of the median
    independently.

    Parameters
    ----------
    samples : array_like, ndim < 3
        Array containing the (multidimensional) samples.
    rate : float, default=0.95
        Fraction of the total mass the interval should contain.
    resolution : int, default=500
        Number of bins used in histogramming the samples.

    Returns
    -------
    :
        Confidence intervals for each component.
    """
    if samples.ndim < 2:
        samples.shape = -1, 1
    assert samples.ndim == 2
    assert 0 <= rate <= 1
    conf_intervals = np.empty((samples.shape[1], 2))
    for j, s in enumerate(samples.T):
        median = np.median(s)
        hist, bdry = np.histogram(s, bins=resolution, density=True)
        median_bin = np.argmax(bdry > median)-1
        cumsum_left = (
            np.cumsum(np.flip(hist[:median_bin+1]))
            - 0.5*hist[median_bin]  # do not count the median bin twice
        )
        cumsum_right = (
            np.cumsum(hist[median_bin:])
            - 0.5*hist[median_bin]  # do not count the median bin twice
        )
        lc = median_bin - np.argmax(cumsum_left >= rate*cumsum_left[-1])
        rc = np.argmax(cumsum_right >= rate*cumsum_right[-1]) + median_bin
        conf_intervals[j] = np.array([bdry[lc], bdry[rc+1]])
    return conf_intervals


def doerfler_marking(values: Union[np.ndarray, List],
                     idx: Optional[np.ndarray] = None,
                     threshold: float = 0.9
                     ) -> Tuple[np.ndarray, np.ndarray, int]:
    """Dörfler marking for arbitrary values.

    Parameters
    ----------
    values : array_like
        Values for the Dörfler marking.
    idx : list of int, optional
        List of indices associated with the entries of `values`.
        If `None`, this is set to ``range(len(values))``.
    threshold : float, default=0.9
        Threshold paramter for Dörfler marking.

    Returns
    -------
    idx_reordered :
        Reordered indices given by `idx`.
        Ordered from largest to smallest value.
    ordered_values :
        Reordered values. Ordered from largest to smallest.
    marker :
        Threshold marker such that
        ``sum(values[:marker]) > threshold * sum(values)``.
    """
    if isinstance(values, list):
        values = np.array(values)
    assert values.size > 0
    if values.ndim < 2:
        values.shape = -1, 1
    assert values.ndim == 2
    if idx is None:
        idx = np.arange(values.shape[0], dtype=int)
    # index list of largest to smalles (absolute) coeff values
    sort = np.flip(np.argsort(np.abs(values), axis=None), axis=0)
    idx_reordered = idx[sort]
    ordered_values = values[sort]
    marker = int(np.argmax(
        np.cumsum(ordered_values, axis=0) > threshold*np.sum(values)))
    return idx_reordered, ordered_values, marker+1


if __name__ == '__main__':
    pass



from . import (
    misc,
    parameter,
    index,
    basis,
    sampler,
    chaos,
)

name = "pythia"
__version__ = "3.0.1"
__author__ = "Nando Farchmin"
__credits__ = ["Physikalisch-Technische Bundesanstalt"]
__maintainer__ = "Nando Farchmin"
__status__ = "Development"



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pythia-uq",
    version="3.0.0",
    author="Nando Farchmin",
    author_email="nando.farchmin@ptb.de",
    description=("Toolbox for non-intrusive functional approximation of data "
                 + "via (sparse) general polynomial chaos."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab1.ptb.de/pythia/pythia",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.20.0",
        "scipy>=1.5.0",
        "psutil>=5.0",
        "sphinx-autodoc-typehints>=1.18.1",
    ],
)



""" Assemble sparse univariate and multivariate basis polynomials.

Build univariate or multivariate normalized basis polynomials depending on
the domain and distribution (and other degrees of freedom) of the parameter(s).

Currently supported are the following distribution types:
    - uniform
    - normal
    - Gamma
    - Beta
"""
from typing import List, Tuple, Optional, Callable, Union
import numpy as np
import scipy.integrate
import scipy.special
import pythia as pt


def univariate_basis(params: List[pt.parameter.Parameter],
                     degs: Union[List[int], Tuple[int], np.ndarray],
                     ) -> List[List[Callable]]:
    """Assemble a univariate polynomial basis.

    Set polynomial basis up to deg for each parameter in `params` according to
    the parameter distribution and area of definition.

    Parameters
    ----------
    params : list of `pythia.parameter.Parameter`
        Parameters to compute univariate basis function for.
    degs : array_like
        Max. degrees of univariate polynomials for each parameter.

    Returns
    -------
    :
        List of normalized univariate polynomials w.r.t. parameter domain and
        distribution up to specified degree for each parameter in `params`.
    """
    basis = []
    param_pdfs = [pt.sampler.assign_sampler(param).pdf for param in params]
    for param, pdf, deg in zip(params, param_pdfs, degs):
        # Set the polynomial basis with corresponding area of support and
        # proper normalization.
        if param.distribution == "uniform":
            polynomials = normalize_polynomial(
                pdf, set_legendre_basis(param, deg), param)
        elif param.distribution == "normal":
            polynomials = normalize_polynomial(
                pdf, set_hermite_basis(param, deg), param)
        elif param.distribution == "gamma":
            polynomials = normalize_polynomial(
                pdf, set_laguerre_basis(param, deg), param)
        elif param.distribution == "beta":
            polynomials = normalize_polynomial(
                pdf, set_jacobi_basis(param, deg), param)
        else:
            raise ValueError(f'Unsupported distribution "{param.distribution}"'
                             f' for {param.name}')
        basis += [polynomials]
    return basis


def multivariate_basis(univariate_bases: List[List[Callable]],
                       indices: np.ndarray,
                       partial: Optional[List[int]] = None,
                       ) -> List[Callable]:
    """Assemble multivariate polynomial basis.

    Set the (partial derivative of the) multivariate (product) polynomial basis
    functions.

    Parameters
    ----------
    univariate_bases : list of list of callable
        Univariate basis functions for parameters. Is called by
        `univariate_bases[paramIdx][deg]()`.
    indices : array_like
        Array of multiindices for multivariate basis functions.
    partial : list of int
        Number of partial derivatives for each dimension. Length is same as
        `univariate_bases`.

    Returns
    -------
    :
        List of multivariate product polynomials with univariate degrees as
        specified in `indices`.
    """
    assert len(univariate_bases) == indices.shape[1]
    if partial is not None:
        assert len(partial) == indices.shape[1]
    basis_list = []
    for index in indices:
        def fun(x: np.ndarray,
                index: Optional[np.ndarray] = index
                ) -> np.ndarray:
            if not 1 <= x.ndim <= 2:
                raise ValueError(f"Wrong ndim '{x.ndim}'")
            if x.ndim == 1:
                x.shape = 1, -1
            if partial is None:
                basis = [univariate_bases[k][mu_k]
                         for k, mu_k in enumerate(index)]
            else:
                basis = [univariate_bases[k][mu_k].deriv(partial[k])
                         for k, mu_k in enumerate(index)]
            return np.prod([basis[k](x[:, k])
                            for k, _ in enumerate(index)], axis=0)
        basis_list += [fun]
    return basis_list


def normalize_polynomial(weight: Callable,
                         basis: List[Callable],
                         param: pt.parameter.Parameter,
                         ) -> List[Callable]:
    """Normalize orthogonal polynomials.

    Normalize a polynomial of an orthogonal system with respect to the scalar
    product

    .. math::
        a(u,v)_\\mathrm{pdf} = \\int u(p) v(p) \\mathrm{pdf}(p) \\mathrm{d}p.

    The normalized polynomial :math:`\\phi_j` for any given polynomial
    :math:`P_j` is given by :math:`\\phi_j = P_j / \\sqrt{c_j}`
    for the constant
    :math:`c_j = \\int \\mathrm{pdf}(p) * P_j(p)^2 \\mathrm{d}p`.

    Parameters
    ----------
    weight : callable
        Probability density function.
    basis : list of `numpy.polynomial.Polynomial`
        Polynomials to normalize w.r.t. weight.
    param : `pythia.parameter.Parameter`
        Parameter used for distribution and domain information.

    Returns
    -------
    :
        List of normalized univariate polynomials.
    """
    cs = np.zeros(len(basis))
    for j, p in enumerate(basis):
        if param.distribution == "normal":
            cs[j] = float(np.math.factorial(j))
        else:
            def integrand(x):
                return weight(x)*p(x)**2
            cs[j], _ = scipy.integrate.quad(
                integrand, param.domain[0], param.domain[1])
    return [p/np.sqrt(c) for c, p in zip(cs, basis)]


def set_legendre_basis(param: pt.parameter.Parameter,
                       deg: int
                       ) -> List[Callable]:
    """Generate list of the Legendre Polynomials.

    Generate the Legendre Polynomials up to certain degree on the interval
    specified by the parameter.

    Parameters
    ----------
    param : `pythia.parameters.Parameter`
        Parameter for basis function. Needs to be uniformly distributed.
    deg : int
        Maximum degree for polynomials.

    Returns
    -------
    :
        List of Legendre polynomials up to (including) degree specified in
        `deg`.
    """
    return [np.polynomial.legendre.Legendre([0]*j+[1], param.domain)
            for j in range(deg+1)]


def set_hermite_basis(param: pt.parameter.Parameter,
                      deg: int
                      ) -> List[Callable]:
    """Generate list of probabilists Hermite polynomials.

    Generate the Hermite Polynomials up to certain degree according to the
    mean and variance of the specified parameter.

    Parameters
    ----------
    param : `pythia.parameters.Parameter`
        Parameter for basis function. Needs to be normal distributed.
    deg : int
        Maximum degree for polynomials.

    Returns
    -------
    :
        List of probabilists Hermite polynomials up to (including) degree
        specified in `deg`.
    """
    assert isinstance(param.mean, (int, float))
    assert isinstance(param.var, (int, float))
    p_list = []
    std = np.sqrt(param.var)
    a = -param.mean / (std * np.sqrt(2))
    b = 1 / (np.sqrt(2) * std)
    shift = np.polynomial.polynomial.Polynomial([a, b])
    for j in range(deg+1):
        p = np.polynomial.hermite.Hermite([0]*j+[1])
        p_list.append(2**(-j/2)*p(shift))
    return p_list


def set_jacobi_basis(param: pt.parameter.Parameter,
                     deg: int
                     ) -> List[Callable]:
    """Generate list of Jacobi polynomials.

    Generate the Jacobi Polynomials up to certain degree on the interval
    and DoFs specified by the parameter.

    .. note::
        The Jacobi polynomials have leading coefficient 1.

    Parameters
    ----------
    param : `pythia.parameters.Parameter`
        Parameter for basis function. Needs to be Beta-distributed.
    deg : int
        Maximum degree for polynomials.

    Returns
    -------
    :
        List of Jacobi polynomials up to (including) degree specified in `deg`.
    """
    assert isinstance(param.alpha, (int, float))
    assert isinstance(param.beta, (int, float))
    p_list = [np.polynomial.polynomial.Polynomial(1)]

    a = pt.misc.shiftCoord(0.0, [-1, 1], param.domain)
    b = pt.misc.shiftCoord(1.0, [-1, 1], param.domain) - a
    shift = np.polynomial.polynomial.Polynomial([a, b])

    for j in range(1, deg+1):
        roots, _ = scipy.special.roots_jacobi(j, param.beta-1, param.alpha-1)
        coeff = np.polynomial.polynomial.polyfromroots(shift(roots))
        p = np.polynomial.polynomial.Polynomial(coeff)
        p_list.append(p)
    return p_list


def set_laguerre_basis(param: pt.parameter.Parameter,
                       deg: int
                       ) -> List[Callable]:
    """Generate list of Leguerre polynomials.

    Generate the generalized Laguerre polynomials up to certain degree on
    the interval and DoFs specified by the parameter.

    Parameters
    ----------
    param : `pythia.parameters.Parameter`
        Parameter for basis function. Needs to be Gamma-distributed.
    deg : int
        Maximum degree for polynomials.

    Returns
    -------
    :
        List of Laguerre polynomials up to (including) degree specified in
        `deg`.
    """
    assert isinstance(param.alpha, (int, float))
    assert isinstance(param.beta, (int, float))
    p_list = [np.polynomial.polynomial.Polynomial(1)]
    shift = np.polynomial.polynomial.Polynomial(
        [param.domain[0], 1/param.beta])
    for j in range(1, deg+1):
        roots, _ = scipy.special.roots_genlaguerre(j, param.alpha-1)
        coeff = np.polynomial.polynomial.polyfromroots(shift(roots))
        p = np.polynomial.polynomial.Polynomial(coeff)
        p_list.append(p)
    return p_list


if __name__ == "__main__":
    pass



""" PyThia classes containing Parameter information. """
from typing import List, Tuple, Optional, Union
from dataclasses import dataclass
import numpy as np


@dataclass
class Parameter:
    """ Class used for stochasic parameters.

    Parameters
    ----------
    index : int
        Enumeration index of the parameter.
    name : str
        Parameter name.
    domain : array_like
        Supported domain of the parameter distribution.
    distribution : str
        Distribution identifier of the parameter.
    mean : float, default=None
        Mean of parameter probability.
    var : float, default=None
        Variance of parameter probability.
    alpha : float, default=None
        Alpha value of Beta and Gamma distribution.
    beta : float, default=None
        Beta value of Beta and Gamma distribution.
    """
    index: int
    name: str
    domain: Union[List, Tuple, np.ndarray]
    distribution: str
    mean: Optional[float] = None
    var: Optional[float] = None
    alpha: Optional[float] = None
    beta: Optional[float] = None

    def __post_init__(self) -> None:
        """Sanity checks."""
        assert isinstance(self.index, int)
        assert isinstance(self.name, str)
        assert isinstance(self.distribution, str)

        self.domain = np.array(self.domain)
        assert self.domain.shape == (2,)

        if self.distribution == "uniform":
            assert self.mean is None
            assert self.var is None
            assert self.alpha is None
            assert self.beta is None
        elif self.distribution == "normal":
            assert isinstance(self.mean, (float, int))
            assert isinstance(self.var, (float, int))
            assert self.var > 0
            assert self.alpha is None
            assert self.beta is None
        elif self.distribution in ["gamma", "beta"]:
            assert self.mean is None
            assert self.var is None
            assert isinstance(self.alpha, (float, int))
            assert isinstance(self.beta, (float, int))
            assert self.alpha >= 0
            assert self.beta >= 0
        else:
            raise ValueError(f"Unknown distribution: '{self.distribution}'")
